// NatNetSample.cpp : Defines the entry point for the application.
//

#define WIN32_LEAN_AND_MEAN		// Exclude rarely-used stuff from Windows headers

#include <windows.h>
#include <stdlib.h>
#include <malloc.h>
#include <memory.h>
#include <tchar.h>
#include <math.h>
#include <stdio.h>

#include <GL/gl.h>
#include <GL/glu.h>
#include "resource.h"

HINSTANCE g_hInst;
HGLRC g_hRenderContext = NULL;

BOOL InitInstance(HINSTANCE, int);
LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK NatNetDlgProc(HWND, UINT, WPARAM, LPARAM);
void Render();


//NatNet SDK
#include "NatNetTypes.h"
#include "NatNetClient.h"

#define MAX_MARKER_COUNT 100
#define MAX_RIGIDBODY_COUNT 200
NatNetClient natnetClient;
int markerCount = 0;
float markerPos[3*MAX_MARKER_COUNT];
int rigidbodyCount = 0;
float rigidbody[7*MAX_RIGIDBODY_COUNT];
bool render = true;
void __cdecl DataHandler(sFrameOfMocapData* data, void* pUserData);			// receives data from the server
void __cdecl MessageHandler(int msgType, char* msg);		// receives NatNet error mesages
bool InitNatNet(LPSTR szIPAddress, LPSTR szServerIPAddress);
bool ParseRigidBodyDescription(sDataDescriptions* pDataDefs);

float g_unitConversion = 1.0f;

int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                     LPSTR lpCmdLine, int nCmdShow)
{
  WNDCLASSEX wcex;
  wcex.cbSize = sizeof(WNDCLASSEX); 
  wcex.style			= CS_HREDRAW | CS_VREDRAW;
  wcex.lpfnWndProc	= (WNDPROC)WndProc;
  wcex.cbClsExtra		= 0;
  wcex.cbWndExtra		= 0;
  wcex.hInstance		= hInstance;
  wcex.hIcon			= NULL;
  wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
  wcex.hbrBackground	= (HBRUSH)(COLOR_WINDOW+1);
  wcex.lpszMenuName	= (LPCSTR)IDC_NATNETSAMPLE;
  wcex.lpszClassName	= "NATNETSAMPLE";
  wcex.hIconSm		= NULL;
  RegisterClassEx(&wcex);

  if (!InitInstance (hInstance, nCmdShow)) 
    return FALSE;

  MSG msg;

  while(TRUE)
  {
    if(PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
    {
      if(!GetMessage(&msg, NULL, 0, 0))
        break;
      TranslateMessage(&msg);
      DispatchMessage(&msg);
    }
    else
    {
      if(render)
      {
        HDC hDC = GetDC(msg.hwnd);
        if(hDC)
        {
          wglMakeCurrent(hDC, g_hRenderContext);
          Render();
          SwapBuffers(hDC);
          wglMakeCurrent(0, 0);
        }
        ReleaseDC(msg.hwnd, hDC);
      }
    }
  }

  return msg.wParam;
}


BOOL InitInstance(HINSTANCE hInstance, int nCmdShow)
{
  g_hInst = hInstance;

  HWND hWnd = CreateWindow("NATNETSAMPLE", "NatNet Sample", WS_OVERLAPPEDWINDOW,
  CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, NULL, NULL, hInstance, NULL);

  if (!hWnd)
    return FALSE;

  // Define pixel format
  PIXELFORMATDESCRIPTOR pfd;
  int nPixelFormat;	
  memset(&pfd, NULL, sizeof(pfd));    
  pfd.nSize      = sizeof(pfd);
  pfd.nVersion   = 1;
  pfd.dwFlags    = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
  pfd.iPixelType = PFD_TYPE_RGBA;
  pfd.cColorBits = 24;
  pfd.cDepthBits = 16;
  pfd.iLayerType = PFD_MAIN_PLANE;

  // Set pixel format
  HDC hDC = GetDC(hWnd);
  nPixelFormat = ChoosePixelFormat(hDC, &pfd);
  SetPixelFormat(hDC, nPixelFormat, &pfd);

  g_hRenderContext = wglCreateContext(hDC);
  wglMakeCurrent(hDC, g_hRenderContext);

  glClearColor(0.000f, 0.000f, 0.000f, 1.0f);

  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHTING);
  glEnable(GL_DEPTH_TEST);

#if 0
  static GLfloat glfMatAmbient[] = {0.000f, 0.450f, 1.000f, 1.0f};
  static GLfloat glfMatDiffuse[] = {0.000f, 0.000f, 0.580f, 1.0f};
  static GLfloat glfMatSpecular[]= {1.000f, 1.000f, 1.000f, 1.0f};
  static GLfloat glfMatEmission[]= {0.000f, 0.000f, 0.000f, 1.0f};
  static GLfloat fShininess = 128.000f;

  glMaterialfv(GL_FRONT, GL_AMBIENT,  glfMatAmbient);
  glMaterialfv(GL_FRONT, GL_DIFFUSE,  glfMatDiffuse);
  glMaterialfv(GL_FRONT, GL_SPECULAR, glfMatSpecular);
  glMaterialfv(GL_FRONT, GL_EMISSION, glfMatEmission);
  glMaterialf(GL_FRONT, GL_SHININESS, fShininess);
#endif

  wglMakeCurrent(0, 0);
  ReleaseDC(hWnd, hDC);

  ShowWindow(hWnd, nCmdShow);
  UpdateWindow(hWnd);

  return TRUE;
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
  int wmId, wmEvent;
  PAINTSTRUCT ps;
  HDC hdc;

  switch (message) 
  {
  case WM_COMMAND:
    wmId    = LOWORD(wParam); 
    wmEvent = HIWORD(wParam); 
    // Parse the menu selections:
    switch (wmId)
    {
    case IDM_CONNECT:
      DialogBox(g_hInst, (LPCTSTR)IDD_NATNET, hWnd, (DLGPROC)NatNetDlgProc);
      break;
    case IDM_EXIT:
      DestroyWindow(hWnd);
      break;
    default:
      return DefWindowProc(hWnd, message, wParam, lParam);
    }
    break;
  case WM_PAINT:
    hdc = BeginPaint(hWnd, &ps);
    wglMakeCurrent(hdc, g_hRenderContext);
    Render();
    SwapBuffers(hdc);
    wglMakeCurrent(0, 0);
    EndPaint(hWnd, &ps);
    break;
  case WM_SIZE:
    {
    int cx = LOWORD(lParam), cy = HIWORD(lParam);
    if(cx==0 || cy ==0 || hWnd==NULL)
      break;

    GLfloat fFovy  = 30.0f; // Field-of-view
    GLfloat fZNear = 1.0f;  // Near clipping plane
    GLfloat fZFar = 10000.0f;  // Far clipping plane

    HDC hDC = GetDC(hWnd);
    wglMakeCurrent(hDC, g_hRenderContext);

    // Calculate viewport aspect
    RECT rv;
    GetClientRect(hWnd, &rv);
    GLfloat fAspect = (GLfloat)(rv.right-rv.left) / (GLfloat)(rv.bottom-rv.top);

    // Define viewport
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(fFovy, fAspect, fZNear, fZFar);
    glViewport(rv.left, rv.top, rv.right-rv.left, rv.bottom-rv.top);
    glMatrixMode(GL_MODELVIEW);

    wglMakeCurrent(0, 0);
    ReleaseDC(hWnd, hDC);
    }
    break;
  case WM_DESTROY:
    {
    HDC hDC = GetDC(hWnd);
    wglMakeCurrent(hDC, g_hRenderContext);
  	natnetClient.Uninitialize();
    wglMakeCurrent(0, 0);
    wglDeleteContext(g_hRenderContext);
    ReleaseDC(hWnd, hDC);
    PostQuitMessage(0);
    }
    break;
  default:
    return DefWindowProc(hWnd, message, wParam, lParam);
  }
  return 0;
}

#define X .5257311f 
#define Z .8506508f

static GLfloat vdata[12][3] = {    
    {-X, 0.0, Z}, {X, 0.0, Z}, {-X, 0.0, -Z}, {X, 0.0, -Z},    
    {0.0, Z, X}, {0.0, Z, -X}, {0.0, -Z, X}, {0.0, -Z, -X},    
    {Z, X, 0.0}, {-Z, X, 0.0}, {Z, -X, 0.0}, {-Z, -X, 0.0} 
};
static GLuint tindices[20][3] = { 
    {0,4,1}, {0,9,4}, {9,5,4}, {4,5,8}, {4,8,1},    
    {8,10,1}, {8,3,10}, {5,3,8}, {5,2,3}, {2,7,3},    
    {7,10,3}, {7,6,10}, {7,11,6}, {11,0,6}, {0,1,6}, 
    {6,1,10}, {9,0,11}, {9,11,2}, {9,2,5}, {7,2,11} };

void normalize(GLfloat *a) {
    GLfloat d=sqrt(a[0]*a[0]+a[1]*a[1]+a[2]*a[2]);
    a[0]/=d; a[1]/=d; a[2]/=d;
}

void drawtri(GLfloat *a, GLfloat *b, GLfloat *c, int div, float r) {
    if (div<=0) {
        glNormal3fv(a); glVertex3f(a[0]*r, a[1]*r, a[2]*r);
        glNormal3fv(b); glVertex3f(b[0]*r, b[1]*r, b[2]*r);
        glNormal3fv(c); glVertex3f(c[0]*r, c[1]*r, c[2]*r);
    } else {
        GLfloat ab[3], ac[3], bc[3];
        for (int i=0;i<3;i++) {
            ab[i]=(a[i]+b[i])/2;
            ac[i]=(a[i]+c[i])/2;
            bc[i]=(b[i]+c[i])/2;
        }
        normalize(ab); normalize(ac); normalize(bc);
        drawtri(a, ab, ac, div-1, r);
        drawtri(b, bc, ab, div-1, r);
        drawtri(c, ac, bc, div-1, r);
        drawtri(ab, bc, ac, div-1, r);  
    }  
}

void drawsphere(int ndiv, float radius=1.0) {
    glBegin(GL_TRIANGLES);
    for (int i=0;i<20;i++)
        drawtri(vdata[tindices[i][0]], vdata[tindices[i][1]], vdata[tindices[i][2]], ndiv, radius);
    glEnd();
}

void quatToMatrix(GLfloat *q, GLfloat *m)
{
  m[0] = 1-2*q[1]*q[1]-2*q[2]*q[2]; m[1] = 2*q[0]*q[1]-2*q[3]*q[2];   m[2] = 2*q[0]*q[2]+2*q[3]*q[1];
  m[3] = 2*q[0]*q[1]+2*q[3]*q[2];   m[4] = 1-2*q[0]*q[0]-2*q[2]*q[2]; m[5] = 2*q[1]*q[2]-2*q[3]*q[0];
  m[6] = 2*q[0]*q[2]-2*q[3]*q[1];   m[7] = 2*q[1]*q[2]+2*q[3]*q[0];   m[8] = 1-2*q[0]*q[0]-2*q[1]*q[1];
}

void vec3MatrixMult(GLfloat *v, GLfloat *m)
{
  GLfloat x = v[0]*m[0]+v[1]*m[3]+v[2]*m[6];
  GLfloat y = v[0]*m[1]+v[1]*m[4]+v[2]*m[7];
  GLfloat z = v[0]*m[2]+v[1]*m[5]+v[2]*m[8];
  v[0] = x;
  v[1] = y;
  v[2] = z;
}

void drawbox(GLfloat x, GLfloat y, GLfloat z, GLfloat qx, GLfloat qy, GLfloat qz, GLfloat qw)
{
  GLfloat s = 50;
  GLfloat m[9], q[] = {qx,qy,qz,qw};
  quatToMatrix(q, m);
  GLfloat p1[] = {s,s,s}, p2[] = {-s,s,s}, p3[] = {-s,-s,s}, p4[] = {s,-s,s};
  GLfloat p5[] = {s,s,-s}, p6[] = {-s,s,-s}, p7[] = {-s,-s,-s}, p8[] = {s,-s,-s};
  vec3MatrixMult(p1, m);
  vec3MatrixMult(p2, m);
  vec3MatrixMult(p3, m);
  vec3MatrixMult(p4, m);
  vec3MatrixMult(p5, m);
  vec3MatrixMult(p6, m);
  vec3MatrixMult(p7, m);
  vec3MatrixMult(p8, m);
  p1[0] += x; p1[1] += y; p1[2] += z;
  p2[0] += x; p2[1] += y; p2[2] += z;
  p3[0] += x; p3[1] += y; p3[2] += z;
  p4[0] += x; p4[1] += y; p4[2] += z;
  p5[0] += x; p5[1] += y; p5[2] += z;
  p6[0] += x; p6[1] += y; p6[2] += z;
  p7[0] += x; p7[1] += y; p7[2] += z;
  p8[0] += x; p8[1] += y; p8[2] += z;

  GLfloat n[3];

  glBegin(GL_POLYGON);
  n[0] = 0; n[1] = 0; n[2] = 1;
  vec3MatrixMult(n, m);
  glNormal3f(n[0], n[1], n[2]);
  glVertex3f(p1[0],p1[1],p1[2]);
  glVertex3f(p2[0],p2[1],p2[2]);
  glVertex3f(p3[0],p3[1],p3[2]);
  glVertex3f(p4[0],p4[1],p4[2]);
  glEnd();

  glBegin(GL_POLYGON);
  n[0] = 0; n[1] = 0; n[2] = -1;
  vec3MatrixMult(n, m);
  glNormal3f(n[0], n[1], n[2]);
  glVertex3f(p5[0],p5[1],p5[2]);
  glVertex3f(p8[0],p8[1],p8[2]);
  glVertex3f(p7[0],p7[1],p7[2]);
  glVertex3f(p6[0],p6[1],p6[2]);
  glEnd();

  glBegin(GL_POLYGON);
  n[0] = -1; n[1] = 0; n[2] = 0;
  vec3MatrixMult(n, m);
  glNormal3f(n[0], n[1], n[2]);
  glVertex3f(p2[0],p2[1],p2[2]);
  glVertex3f(p6[0],p6[1],p6[2]);
  glVertex3f(p7[0],p7[1],p7[2]);
  glVertex3f(p3[0],p3[1],p3[2]);
  glEnd();

  glBegin(GL_POLYGON);
  n[0] = 1; n[1] = 0; n[2] = 0;
  vec3MatrixMult(n, m);
  glNormal3f(n[0], n[1], n[2]);
  glVertex3f(p1[0],p1[1],p1[2]);
  glVertex3f(p4[0],p4[1],p4[2]);
  glVertex3f(p8[0],p8[1],p8[2]);
  glVertex3f(p5[0],p5[1],p5[2]);
  glEnd();

  glBegin(GL_POLYGON);
  n[0] = 0; n[1] = 1; n[2] = 0;
  vec3MatrixMult(n, m);
  glNormal3f(n[0], n[1], n[2]);
  glVertex3f(p6[0],p6[1],p6[2]);
  glVertex3f(p2[0],p2[1],p2[2]);
  glVertex3f(p1[0],p1[1],p1[2]);
  glVertex3f(p5[0],p5[1],p5[2]);
  glEnd();

  glBegin(GL_POLYGON);
  n[0] = 0; n[1] = -1; n[2] = 0;
  vec3MatrixMult(n, m);
  glNormal3f(n[0], n[1], n[2]);
  glVertex3f(p7[0],p7[1],p7[2]);
  glVertex3f(p8[0],p8[1],p8[2]);
  glVertex3f(p4[0],p4[1],p4[2]);
  glVertex3f(p3[0],p3[1],p3[2]);
  glEnd();
}

void Render()
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); // Clear buffers
  glLoadIdentity(); // Load identity matrix

  GLfloat glfLight[] = {-4.0f, 4.0f, 4.0f, 0.0f};
  glLightfv(GL_LIGHT0, GL_POSITION, glfLight);

  // Position and rotate the camera
  glTranslatef(0.0f, -1000.0f, -5000.0f);	

  //draw axis
  glBegin(GL_LINES);
	glColor3f(1.0f, 0.0f, 0.0f);
  glVertex3f(0,0,0);
  glVertex3f(300,0,0);

	glColor3f(0.0f, 1.0f, 0.0f);
  glVertex3f(0,0,0);
  glVertex3f(0,300,0);

	glColor3f(0.0f, 0.0f, 1.0f);
  glVertex3f(0,0,0);
  glVertex3f(0,0,300);
  glEnd();

  int i;
  for (i=0;i<rigidbodyCount;i++)
  {
    GLfloat x = rigidbody[7*i], y = rigidbody[7*i+1], z = rigidbody[7*i+2];
	float scale = 100.0f;
	x *= g_unitConversion;	// convert to mm
	y *= g_unitConversion;
	z *= g_unitConversion;
    GLfloat qx = rigidbody[7*i+3], qy = rigidbody[7*i+4], qz = rigidbody[7*i+5], qw = rigidbody[7*i+6];
    drawbox(x,y,z,qx,qy,qz,qw);
  }

  for (i=0;i<markerCount;i++)
  {
	  glPushMatrix();
	  glTranslatef(markerPos[3*i], markerPos[3*i+1], markerPos[3*i+2]);
    drawsphere(1,20);
    glPopMatrix();
  }

  glFlush();

  render = false;
}

LRESULT CALLBACK NatNetDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
  switch (message)
  {
  case WM_INITDIALOG:
    SetDlgItemText(hDlg, IDC_EDIT1, "127");
    SetDlgItemText(hDlg, IDC_EDIT2, "0");
    SetDlgItemText(hDlg, IDC_EDIT3, "0");
    SetDlgItemText(hDlg, IDC_EDIT4, "1");
    SetDlgItemText(hDlg, IDC_EDIT5, "127");
    SetDlgItemText(hDlg, IDC_EDIT6, "0");
    SetDlgItemText(hDlg, IDC_EDIT7, "0");
    SetDlgItemText(hDlg, IDC_EDIT8, "1");
    return TRUE;

  case WM_COMMAND:
    switch (LOWORD(wParam))
    {
    case IDC_CONNECT:
      {
      char szMyIPAddress[30], szServerIPAddress[30];
      char ip1[5], ip2[5], ip3[5], ip4[5];
      GetDlgItemText(hDlg, IDC_EDIT1, ip1, 4);
      GetDlgItemText(hDlg, IDC_EDIT2, ip2, 4);
      GetDlgItemText(hDlg, IDC_EDIT3, ip3, 4);
      GetDlgItemText(hDlg, IDC_EDIT4, ip4, 4);
      sprintf_s(szMyIPAddress, 30, "%s.%s.%s.%s", ip1,ip2,ip3,ip4);

      GetDlgItemText(hDlg, IDC_EDIT5, ip1, 4);
      GetDlgItemText(hDlg, IDC_EDIT6, ip2, 4);
      GetDlgItemText(hDlg, IDC_EDIT7, ip3, 4);
      GetDlgItemText(hDlg, IDC_EDIT8, ip4, 4);
      sprintf_s(szServerIPAddress, 30, "%s.%s.%s.%s", ip1,ip2,ip3,ip4);

      if (InitNatNet(szMyIPAddress, szServerIPAddress) == FALSE)
      {
        natnetClient.Uninitialize();
        MessageBox(hDlg, "Failed to connect", "", MB_OK);
      }
      }
    case IDOK:
    case IDCANCEL:
      EndDialog(hDlg, LOWORD(wParam));
      return TRUE;
    }
  }
  return FALSE;
}


bool InitNatNet(LPSTR szIPAddress, LPSTR szServerIPAddress)
{
  unsigned char ver[4];
  natnetClient.NatNetVersion(ver);

  // Set callback handlers
  natnetClient.SetMessageCallback(MessageHandler);
  natnetClient.SetVerbosityLevel(Verbosity_None);
  natnetClient.SetDataCallback( DataHandler );	// this function will receive data from the server

  int retCode = natnetClient.Initialize(szIPAddress, szServerIPAddress);
  if (retCode != ErrorCode_OK)
  {
    //Unable to connect to server.
    return FALSE;
  }
  else
  {
    // print server info
    sServerDescription ServerDescription;
    memset(&ServerDescription, 0, sizeof(ServerDescription));
    natnetClient.GetServerDescription(&ServerDescription);
    if(!ServerDescription.HostPresent)
    {
      //Unable to connect to server. Host not present
      return FALSE;
    }
  }

  // Retrieve RigidBody description from server
  sDataDescriptions* pDataDefs = NULL;
  int nBodies = natnetClient.GetDataDescriptions(&pDataDefs);
  if (ParseRigidBodyDescription(pDataDefs) == false)
  {
    //Unable to retrieve RigidBody description
    //return FALSE;
  }

  // example of NatNet general message passing
  void* response;
  int nBytes;
  retCode = natnetClient.SendMessageAndWait("UnitsToMillimeters", &response, &nBytes);
  if (retCode == ErrorCode_OK)
  {
    g_unitConversion = *(float*)response;
  }


  return TRUE;
}

bool ParseRigidBodyDescription(sDataDescriptions* pDataDefs)
{
  if (pDataDefs == NULL || pDataDefs->nDataDescriptions <= 0)
    return false;

  for (int i=0,j=0;i<pDataDefs->nDataDescriptions;i++)
  {
    if (pDataDefs->arrDataDescriptions[i].type != Descriptor_RigidBody)
      continue;
    
    sRigidBodyDescription *pRB = pDataDefs->arrDataDescriptions[i].Data.RigidBodyDescription;
    //construct skeleton hierarchy
    int ID        = pRB->ID;
    int parentID  = pRB->parentID;
    float offsetx = pRB->offsetx;
    float offsety = pRB->offsety;
    float offsetz = pRB->offsetz;
  }

  return true;
}

void MessageHandler(int msgType, char* msg)
{
  //	printf("\n[SampleClient] Message received: %s\n", msg);
}

void DataHandler(sFrameOfMocapData* data, void* pUserData)
{
  int i;
  int mcount = min(MAX_MARKER_COUNT,data->MocapData->nMarkers);
 	for(i=0;i<mcount;i++)
	{
    markerPos[3*i]   = data->MocapData->Markers[i][0]; //x pos
    markerPos[3*i+1] = data->MocapData->Markers[i][1]; //y pos
    markerPos[3*i+2] = data->MocapData->Markers[i][2]; //z pos
	}
  markerCount = mcount;

  int rbcount = min(MAX_RIGIDBODY_COUNT, data->nRigidBodies);
  for (i=0;i<rbcount;i++)
  {
    rigidbody[7*i]   = data->RigidBodies[i].x;  //x pos
    rigidbody[7*i+1] = data->RigidBodies[i].y;  //y pos
    rigidbody[7*i+2] = data->RigidBodies[i].z;  //z pos
    rigidbody[7*i+3] = data->RigidBodies[i].qx; //quaternion x
    rigidbody[7*i+4] = data->RigidBodies[i].qy; //quaternion y
    rigidbody[7*i+5] = data->RigidBodies[i].qz; //quaternion z
    rigidbody[7*i+6] = data->RigidBodies[i].qw; //quaternion w
  }
  rigidbodyCount = rbcount;

  render = true;
}
